from ._zermelo_api import zermelo
from ._zermelo_collection import from_zermelo_dict
from dataclasses import dataclass, InitVar, field
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

@dataclass
class Appointment:
    id: int
    appointmentInstance: int = 0
    start: int = 0
    end: int = 0
    startTimeSlot: int = 0
    endTimeSlot: int = 0
    branch: int = 0
    type: str = "unknown"
    groupsInDepartments: list[int] = field(default_factory=list)
    locationsOfBranch: list[int] = field(default_factory=list)
    locations: list[str] = field(default_factory=list)
    optional: bool = False
    valid: bool = False
    cancelled: bool = False
    cancelledReason: str = ""
    modified: bool = False
    teacherChanged: bool = False
    groupChanged: bool = False
    locationChanged: bool = False
    timeChanged: bool = False
    moved: bool = False
    created: int = 0
    hidden: bool = False
    commonSchedule: bool = False
    ignoreSubstitutions: bool = False
    changeDescription: str = ""
    schedulerRemark: str = ""
    capacity: int = 0
    content: str = ""
    lastModified: int = 0
    new: bool = True
    choosableInDepartments: list[int] = field(default_factory=list)
    choosableInDepartmentCodes: list[str] = field(default_factory=list)
    courses: list[int] = field(default_factory=list)
    alternativeSubject: str = ""
    onlineStudents: list[str] = field(default_factory=list)
    extraStudentSource: str = ""
    appointmentLastModified: int = 0
    onlineLocationUrl: str = ""
    remark: str = ""
    capacityManually: bool = False
    teachingTimeManually: bool = False
    teachingTime: int = 0
    expectedStudentCount: int = 0
    expectedStudentCountOnline: int = 0
    availableSpace: int = 0
    udmUUID: str = ""
    creator: str = ""
    subjects: list[str] = field(default_factory=list)
    teachers: list[str] = field(default_factory=list)
    onlineTeachers: list[str] = field(default_factory=list)
    students: list[str] = field(default_factory=list)

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __hash__(self):
        return hash((self.id, self.lastModified))


async def get_appointments(query: str) -> list[Appointment]:
    try:
        status, data = await zermelo.getData(query)
        if isinstance(data, Exception):
            logger.error(f"status: {status}")
            raise Exception(data)
        if status != 200:
            logger.error(f"error: {status}")
            raise Exception(data)
        if not isinstance(data, list):
            raise Exception(f"{type(data)} is not a list")
        return [from_zermelo_dict(Appointment, row) for row in data]
    except Exception as e:
        logger.exception(e)
        logger.error(query)
        return []


async def get_location_appointments(
    lok: int, schoolyear: int, **kwargs
) -> list[Appointment]:
    query = f"appointments/?locationsOfBranch={lok}&schoolInSchoolYear={schoolyear}"
    for key, val in kwargs.items():
        query += f"&{key}={val}"
    logger.debug(query)
    return await get_appointments(query)


async def get_user_appointments(user: int | str, **kwargs) -> list[Appointment]:
    query = f"appointments/?user={user}"
    for key, val in kwargs.items():
        query += f"&{key}={val}"
    logger.debug(query)
    return await get_appointments(query)


async def get_department_updates(id: int, **kwargs) -> list[Appointment]:
    query = f"appointments/?containsStudentsFromGroupInDepartment={id}"
    for key, val in kwargs.items():
        query += f"&{key}={val}"
    return await get_appointments(query)



from __future__ import annotations
from ._credentials import Credentials
from ._io_json import get_json, post_request
import json
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class NoSchoolNameException(Exception):
    def __init__(self, message: str = ""):
        error = f"no schoolname provided: {message}\n Please import and run installAPI to setup credentials"
        super().__init__(error)


class NoTokenException(Exception):
    def __init__(self, message: str = ""):
        error = f"no token provided: {message}\n Please Please import and run installAPI to setup credentials"
        super().__init__(error)


class WrongCredentialsException(Exception):
    def __init__(self, message: str = ""):
        error = f"wrong token provided: {message}\n Please Please import and run installAPI to setup credentials"
        super().__init__(error)


class ZermeloAPI:
    def __init__(self):
        self.credentials = Credentials()
        try:
            if not self.credentials.schoolname:
                raise NoSchoolNameException("init API")
            self.set_schoolname(self.credentials.schoolname)
            if not self.credentials.token:
                raise NoTokenException("init API")
        except Exception as e:
            logger.exception(e)

    def set_schoolname(self, schoolname: str):
        self.zerurl = f"https://{schoolname}.zportal.nl/api/v3/"

    def update_schoolname(self, schoolname: str):
        if not schoolname:
            raise Exception("No schoolname provided")
        if schoolname == self.credentials.schoolname:
            logger.debug("schoolname is already stored")
            return
        self.credentials.setschoolname(schoolname)
        self.set_schoolname(schoolname)
        self.credentials.settoken("")

    async def login(self, code: str) -> bool:
        token = await self.get_access_token(code)
        return await self.add_token(token)

    async def get_access_token(self, code: str) -> str:
        if not code:
            raise Exception("No Code Provided")
        code = "".join(code.split())
        logger.debug(f"new code {code}")
        url = self.zerurl + f"oauth/token"
        data = {"grant_type": "authorization_code", "code": code}
        return await post_request(url, data)

    async def add_token(self, token: str) -> bool:
        if not token:
            return False
        self.credentials.settoken(token)
        return await self.checkCreds()

    async def checkCreds(self):
        try:
            await self.getName()
            result = True
        except Exception as e:
            logger.error(e)
            result = False
        finally:
            return result

    async def getName(self):
        if not self.credentials.schoolname:
            raise NoSchoolNameException("getName")
        if not self.credentials.token:
            raise NoTokenException("using getName")
        status, data = await self.getData("users/~me", True)
        if status != 200 or type(data) is not list:
            raise WrongCredentialsException("getName")
        logger.debug(f"get name: {data[0]}")
        row = data[0]
        if not row["prefix"]:
            return " ".join([row["firstName"], row["lastName"]])
        else:
            return " ".join([row["firstName"], row["prefix"], row["lastName"]])

    async def getData(
        self, task, from_id=False
    ) -> tuple[int, list[dict] | str | Exception]:
        result = (500, "unknown error")
        request = (
            self.zerurl + task + f"?access_token={self.credentials.token}"
            if from_id
            else self.zerurl + task + f"&access_token={self.credentials.token}"
        )
        logger.debug(request)
        try:
            data1 = await get_json(request)
            json_response = json.loads(data1.decode("utf-8"))
            if json_response:
                json_status = json_response["response"]["status"]
                if json_status == 200:
                    result = (200, json_response["response"]["data"])
                    logger.debug("    **** JSON OK ****")
                else:
                    logger.debug(f"oeps, geen juiste response: {task}")
                    result = (json_status, json_response["response"])
            else:
                logger.error("JSON - response is leeg")
        except Exception as e:
            logger.exception(e)
            result = (500, e)
        finally:
            return result

    async def load_query(self, query: str) -> list[dict]:
        try:
            status, data = await self.getData(query)
            if status != 200 or type(data) is not list:
                raise Exception(f"Error loading data {status}, {data}")
            if not data:
                logger.debug("no data")
            return data
        except Exception as e:
            logger.debug(e)
            return []


zermelo = ZermeloAPI()


async def loadAPI(schoolname: str = "") -> ZermeloAPI:
    try:
        if schoolname:
            zermelo.update_schoolname(schoolname)
        if not await zermelo.checkCreds():
            logger.warning(
                "credentials not set up properly. Please import and run: installAPI()"
            )
    except Exception as e:
        logger.exception(e)
    finally:
        return zermelo



from ._zermelo_collection import ZermeloCollection
from .vakken import Vak
from dataclasses import dataclass, InitVar
import logging

logger = logging.getLogger(__name__)


@dataclass
class Groep:
    id: int
    isMainGroup: bool
    isMentorGroup: bool
    departmentOfBranch: int
    name: str
    extendedName: str


class deelnemers(tuple[list[int], list[str], list[str]]): ...


class Groepen(ZermeloCollection[Groep]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"groupindepartments?schoolInSchoolYear={schoolinschoolyear}"
        super().__init__(Groep, query)

    def get_department_groups(
        self, departmentOfBranch: int, maingroup: bool = False
    ) -> list[Groep]:
        # returns all groups (if it is/isnt stamklas (maingroup))
        return [
            groep
            for groep in self
            if groep.departmentOfBranch == departmentOfBranch
            and groep.isMainGroup == maingroup
        ]

    def get_vakgroepen(self, vak: Vak) -> list[Groep]:
        result = []
        logger.debug(f"finding groep for vak: {vak.subjectCode}")
        for groep in self.get_department_groups(vak.departmentOfBranch):
            if groep in result:
                continue
            if vak.qualifiedCode and vak.qualifiedCode in groep.extendedName:
                logger.debug(
                    f"found {groep.name} with {vak.qualifiedCode} in {groep.extendedName}"
                )
                result.append(groep)
        return result



from ._config import getConfig
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class Credentials:
    def __init__(self):
        self.file = getConfig("creds")
        self.schoolname = ""
        self.token = ""
        self.load()

    def load(self):
        try:
            data = self.file.load()
            self.schoolname = data["schoolname"]
            self.token = data["token"]
        except Exception as e:
            logger.error(e)
        finally:
            logger.debug(f"schoolname: {self.schoolname}")
            logger.debug(f"token: {self.token}")

    def check(self) -> bool:
        if self.schoolname and self.token:
            return True
        return False

    def save(self):
        data = {}
        if self.schoolname:
            data["schoolname"] = self.schoolname
        if self.token:
            data["token"] = self.token
        self.file.save(data)

    def setschoolname(self, schoolname: str):
        logger.debug(f"setting schoolname: {schoolname}")
        self.schoolname = schoolname
        self.save()

    def settoken(self, token: str):
        logger.debug(f"setting token: {token}")
        self.token = token
        self.save()



from ._zermelo_collection import ZermeloCollection
from .vakken import Vak
from .groepen import Groep
from ._time_utils import get_date, delta_week
from dataclasses import dataclass, InitVar, field
import asyncio
import logging

logger = logging.getLogger(__name__)

skip_docs: list[str] = ["stth", "lgverv", "lgst"]


def check_doc_skip(doc: str) -> bool:
    for skip_doc in skip_docs:
        if skip_doc.lower() in doc.lower():
            return True
    return False


def clean_checklist(checklist: list[str]):
    for doc in reversed(checklist):
        if check_doc_skip(doc):
            checklist.remove(doc)
    return checklist


@dataclass
class VakLes:
    id: int
    appointmentInstance: int
    teachers: list[str]
    students: list[str]
    subjects: list[str]
    groups: list[str]
    groupsInDepartments: list[int]
    choosableInDepartmentCodes: list[str]
    valid: bool
    cancelled: bool

    def filter(self, name: str) -> bool:
        logger.debug(f"filtering {self} ({name})")
        if self.cancelled:
            logger.debug("cancelled")
            return False
        if not self.valid:
            logger.debug("invalid")
            return False
        if not len(self.students):
            logger.debug("no students")
            return False
        if not len(self.teachers):
            logger.debug("no teachers")
            return False
        if len(self.students) > 40:
            logger.debug("groep te groot")
            return False
        if not any([name.split(".")[-1] in group for group in self.groups]):
            logger.debug(f"{name} not in {self}")
            return False
        logger.debug("True")
        return True


def clean_docs(docs: list[str]) -> list[str]:
    checklist = list(set(docs))
    clean_checklist(checklist)
    max = 0
    if len(checklist) > 1:
        logger.debug(f"multiple docs: {checklist}")
        for doc in checklist:
            doc_count = docs.count(doc)
            if doc_count > max:
                result = doc
                max = doc_count
        logger.debug(f"result: {result} ({max})")
        return [result]
    return checklist


class LesData(tuple[list[int], list[str], list[str]]):
    def __new__(cls, *U):
        return super(LesData, cls).__new__(cls, tuple(U))


class VakLessen(ZermeloCollection[VakLes]):
    id: InitVar[int] = 0
    code: InitVar[str] = ""
    groupName: str = ""
    start: InitVar[int] = 0
    eind: InitVar[int] = 0

    def __init__(
        self, id: int, code: str, groupName: str = "", start: int = 0, eind: int = 0
    ):
        query = f"appointments/?containsStudentsFromGroupInDepartment={id}&subjects={code}&type=lesson&start={start}&end={eind}&fields=appointmentInstance,id,teachers,students,subjects,groups,groupsInDepartments,choosableInDepartmentCodes,valid,cancelled"
        super().__init__(VakLes, query)
        self.groupName = groupName

    def filter(self) -> list[VakLes]:
        logger.debug(f"filtering {self}")
        return [les for les in self if les.filter(self.groupName)]

    def get_data(self) -> LesData:
        grp_bck = []
        ll_bck = []
        doc_bck = []
        leerlingen = []
        docenten = []
        grp_namen = []
        for les in self.filter():
            if len(les.students) > 40:
                if not grp_namen and (not grp_bck or len(les.groups) < len(grp_bck)):
                    logger.debug("meerdere groepen")
                    grp_bck = les.choosableInDepartmentCodes
                    ll_bck = list(set([llnr for llnr in les.students]))
                    doc_bck = list(set([doc for doc in les.teachers]))
                continue
            [leerlingen.append(llnr) for llnr in les.students if llnr not in leerlingen]
            [docenten.append(doc) for doc in les.teachers]
            [
                grp_namen.append(grp)
                for grp in les.choosableInDepartmentCodes
                if grp not in grp_namen
            ]
        logger.debug(
            f"filtering data for {self} \ngrp_namen: {grp_namen} \ndocenten: {docenten} \nleerlingen: {leerlingen}"
        )
        logger.debug(f"")
        logger.debug(
            f"backupdata: \n grp_bck: {grp_bck}\n doc_bck: {doc_bck} \nll_bck: {ll_bck}"
        )
        if not grp_namen and grp_bck:
            logger.debug(f"result groepen: {grp_bck}")
            grp_namen = grp_bck
        if not docenten and doc_bck:
            logger.debug(f"result docenten: {doc_bck}")
            docenten = doc_bck
        if not leerlingen and ll_bck:
            logger.debug(f"result leerlingen: {ll_bck}")
            leerlingen = ll_bck
        docenten = clean_docs(docenten)
        logger.debug(f"after cleaning docs: {docenten}")
        leerlingen = [int(llnr) for llnr in leerlingen]
        return LesData(leerlingen, docenten, grp_namen)


def create_new_vaklessen(vak: Vak, groep: Groep) -> list[VakLessen]:
    date = get_date()
    result: list[VakLessen] = []
    for x in [0, -1, 1, -2, 2, 3, -3]:
        dweek = x * 4
        starttijd = int(delta_week(date, dweek).timestamp())
        eindtijd = int(delta_week(date, dweek + 4).timestamp())
        result.append(
            VakLessen(
                groep.id,
                vak.subjectCode,
                groep.extendedName,
                starttijd,
                eindtijd,
            )
        )
    return result


async def get_vakgroep_lessen(vak: Vak, groep: Groep) -> VakLessen | None:
    logger.debug(f"getting vakgroep lessen for {vak} and {groep}")
    try:
        result = create_new_vaklessen(vak, groep)
        for vaklessen in result:
            logger.debug(f"init: {vaklessen}")
            await vaklessen._init()
            if not len(vaklessen):
                logger.debug("geen lessen gevonden")
                continue
            lessen = vaklessen.filter()
            logger.debug(f"lessen: {lessen}")
            if len(lessen):
                return vaklessen
        logger.debug("geen valid vaklessen gevonden.")
    except Exception as e:
        logger.error(e)


def check_data(data: LesData, vak: Vak) -> LesData | None:
    logger.debug(f"checking data for: \n  data: {data}\n  vak: {vak}")
    leerlingen, docenten, groep_namen = data
    if len(leerlingen) and len(docenten):
        namen = [
            groepnaam
            for groepnaam in groep_namen
            if vak.departmentOfBranchCode in groepnaam
        ]
        return LesData(leerlingen, docenten, namen)
    return None


async def get_vakgroep_data(vak, groep) -> LesData | None:
    logger.debug(f"getting vakgroep data for: \n  vak: {vak}\n  groep: {groep}")
    vaklessen = await get_vakgroep_lessen(vak, groep)
    if not vaklessen:
        logger.debug("geen lessen")
        return None
    logger.debug(f"vaklessen: {vaklessen}")
    lesdata = vaklessen.get_data()
    return check_data(lesdata, vak)


async def get_groep_data(vak: Vak, groep: Groep) -> tuple[Groep, LesData | None]:
    logger.debug(f"getting data for: \n  vak: {vak}\n  groep: {groep}")
    data = await get_vakgroep_data(vak, groep)
    return (groep, data)



from .src._zermelo_api import ZermeloAPI, loadAPI
from .src.install import installAPI
from .src.schoolyears import SchoolYears, SchoolInSchoolYear
from .src.branches import Branch, Branches, load_branches, load_schools
from .src._time_utils import *
from .src.users import Leerling, Medewerker, Leerlingen, Personeel
from .src.lesgroepen import Lesgroepen, Lesgroep, Groep, Groepen
from .src.leerjaren import Leerjaar, Leerjaren
from .src.appointments import (
    get_user_appointments,
    get_department_updates,
    get_location_appointments,
    Appointment,
)
from .src.vakken import Vakken, Vak
from .src.lokalen import Lokalen, Lokaal
from .src.vakdoclok import VakLoks, VakLok, get_vakloks



from ._zermelo_api import loadAPI
import asyncio
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(filename="example.log", encoding="utf-8", level=logging.DEBUG)


async def main(schoolname: str, code: str) -> bool:
    api = await loadAPI(schoolname)
    try:
        await api.login(code)
        if await api.checkCreds():
            print("token set up correctly")
            return True
    except Exception as e:
        logger.exception(e)
    print("error setting up credentials")
    return False


def installAPI(schoolname: str = "", code: str = ""):
    print("updating credentials")
    while not schoolname:
        schoolname = input("name of school:")

    while not code:
        code = input("code: ")

    if not asyncio.run(main(schoolname, code)):
        installAPI()


if __name__ == "__main__":
    installAPI()



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="zermelo_api_vogk",
    version="1.3.3",
    description="A small module to create a Zermelo accesstoken and put some data from Zermelo in dataclasses",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KlaasVogel/zermelo_api_vogk",
    author="Klaas Vogel",
    author_email="zermelo_api@klaasvogel.nl",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    install_requires=["aiohttp"],
    extras_require={
        "dev": [],
    },
    python_requires=">=3.12",
)



from ._zermelo_collection import ZermeloCollection
from ._time_utils import get_date, datetime
from .schoolyears import SchoolYears, SchoolInSchoolYear, load_schoolyears
from .users import Leerlingen, Personeel
from .leerjaren import Leerjaren
from .groepen import Groepen
from .lesgroepen import Lesgroepen
from .vakken import Vakken
from .lokalen import Lokalen
from .vakdoclok import get_vakloks, VakLoks
from dataclasses import dataclass, field, InitVar
import asyncio
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

@dataclass
class Branch:
    id: int
    schoolInSchoolYear: int
    branch: str
    name: str
    schoolYear: int
    date: datetime = datetime.now()
    leerlingen: Leerlingen = field(default_factory=Leerlingen)
    personeel: Personeel = field(default_factory=Personeel)
    leerjaren: Leerjaren = field(default_factory=Leerjaren)
    vakken: Vakken = field(default_factory=Vakken)
    groepen: Groepen = field(default_factory=Groepen)
    lokalen: Lokalen = field(default_factory=Lokalen)

    def __post_init__(self):
        logger.info(f"*** loading branch: {self.name} ***")
        self.leerlingen = Leerlingen(self.schoolInSchoolYear)
        self.personeel = Personeel(self.schoolInSchoolYear)
        self.leerjaren = Leerjaren(self.schoolInSchoolYear)
        self.groepen = Groepen(self.schoolInSchoolYear)
        self.vakken = Vakken(self.schoolInSchoolYear)
        self.lokalen = Lokalen(self.schoolInSchoolYear)

    async def _init(self):
        attrs = ["leerlingen", "personeel", "leerjaren", "groepen", "vakken", "lokalen"]
        await asyncio.gather(
            *[getattr(self, name)._init() for name in attrs], return_exceptions=False
        )

    async def find_lesgroepen(self) -> Lesgroepen | None:
        if self.leerlingen and self.personeel:
            return await Lesgroepen.create(
                self.leerjaren,
                self.vakken,
                self.groepen,
                self.leerlingen,
                self.personeel,
            )

    async def get_vak_doc_loks(self) -> VakLoks:
        start = int(self.date.timestamp())
        eind = start + 28 * 24 * 3600
        return await get_vakloks(self.id, start, eind)


class Branches(ZermeloCollection[Branch]):

    def __init__(self):
        super().__init__(Branch, "branchesofschools/")

    async def _init(self, schoolyears: SchoolYears, datestring: str = ""):
        logger.debug("init branches")
        date = get_date(datestring)
        logger.debug(f"date: {date}")
        await asyncio.gather(
            *[self.load_from_schoolyear(sy, date) for sy in schoolyears]
        )
        await asyncio.gather(*[branch._init() for branch in self])
        logger.info(self)

    async def load_from_schoolyear(self, sy: SchoolInSchoolYear, date: datetime):
        query = f"branchesofschools/?schoolInSchoolYear={sy.id}"
        await self.load_collection(query, date=date)

    def __str__(self):
        return "Branches(" + ", ".join([br.name for br in self]) + ")"

    def get(self, name: str) -> Branch | None:
        for branch in self:
            if (
                name.lower() in branch.branch.lower()
                or branch.branch.lower() in name.lower()
            ):
                return branch
        else:
            logger.error(f"NO Branch found for {name}")


async def load_branches(schoolname: str, date: str = "") -> Branches | None:
    try:
        data = await load_schools(schoolname, date)
        if not data:
            raise Exception("No branches found")
        _, branches = data
        return branches
    except Exception as e:
        logger.error(e)


async def load_schools(
    schoolname: str, date: str = ""
) -> tuple[SchoolYears, Branches] | None:
    try:
        logger.debug("loading schools:")
        schoolyears = await load_schoolyears(schoolname, date)
        logger.debug(schoolyears)
        branches = Branches()
        if not schoolyears:
            return None
        await branches._init(schoolyears, date)
        return (schoolyears, branches)
    except Exception as e:
        logger.error(e)
